/*
 * Get random numbers from a hardware random number generator.
 * 
 * 
 * 3 kinds of operating system support for the RNG are
 * supported.
 * 
 * 1) If the OS provides a "file" that can be opened with
 * fopen such that each byte read is a random number from the device.
 * RNGDRIVER option should be set to the filespec in config.sys
 * 
 * 2) a bus RNG under MSDOS. get each random byte with the
 * "IN" instruction from the port. RNGPORT must be defined in config.txt
 * as the port to read. If it is necessary to wait between random
 * numbers to insure independance, then RANDHARDWAIT (defined on compile,
 * default=48) is the number of clicks to wait for.
 * 
 * 3) a bus RNG which is accessed using OS/2 's TESTCFG driver.
 * as above the config.txt option, RNGPORT is the port to read.
 * 
 */

#include <stdio.h>
#include <stdlib.h>
#include "hardrand.h"

#if defined(OS2) || defined(__OS2__)
#define INCL_DOSFILEMGR
#define INCL_DOSDEVICES
#include <os2.h>
#endif

#if defined(MSDOS) || defined(__MSDOS__)
#include <conio.h>
#ifndef RARNHARDWAIT
#define RANDHARDWAIT (0x48)
#endif
/* outp() and inp() works only for Microsoft C for IBM PC or AT */
/* timer0 on 8253-5 on IBM PC or AT tics every .84 usec. */
#define timer0		0x40	/* 8253 timer 0 port */
#define timercntl	0x43	/* 8253 control register */
#define timer0rwl	0x00	/* read lo/hi bytes of cntr 2 with latch */
#define timer0rnl	0x30	/* read lo/hi bytes of cntr 2 w/o latch */
#endif

#include	<stdio.h>	/* for putchar() and printf() */

#include "pgp.h"
#include "exitpgp.h"
#include "language.h"

/* Possible error exit codes - not all of these are used.  Note that we
   don't use the ANSI EXIT_SUCCESS and EXIT_FAILURE.  To make things
   easier for compilers which don't support enum we use #defines */

#define EXIT_OK				0
#define INVALID_FILE_ERROR		1
#define FILE_NOT_FOUND_ERROR		2
#define UNKNOWN_FILE_ERROR		3
#define NO_BATCH			4
#define BAD_ARG_ERROR			5
#define INTERRUPT			6
#define OUT_OF_MEM			7

/*
*  Perform a crude test to see if RNG is actually working!
*  return 0 if hardrand appears to be random (crude test)
*  1 otherwise
*/
static int RNGgood(void)
{
	/* table list is such that list[i] is a count of bits on in i. */
	const byte list[16]= {0,1,1,2,1,2,2,3,1,2, 2, 3, 2, 3, 3, 4};
	/*                    0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15 */
	const int  test_bytes=100;
	int i;			/* index */
	byte current;		/* current byte */
	int bitcount;		/* count of bits on */
	int shouldbitcount;	/* number of bits that should be on */
	int delta;		/* actual differs from shouldbitcount by this */
	int tw;			/* permmitted slop */
	tw = (test_bytes * 8) / 5 ;	/* permit slop 20% */
	/* half of bits should be on */
	shouldbitcount = (test_bytes * 4);   /* 4= 8/2 */

	/* count bits on */
	bitcount=0;
	for(i=0;i<test_bytes;i++)
	{
		/* use to values xored this will detect constat values */
		current = hardrand() ^ hardrand();
		/* count bits in current */
		/* number of bits in 1st 4 bits */
		bitcount += list [current & 0xf ];
		/* number of bits in 2nd 4 bits */
		bitcount += list [(current >> 4) & 0xf ];
	};
	/* bitcount should be approx 50% of test_bytes * 8 */
	delta = shouldbitcount - bitcount;
	/* difference from expected value */
	/* and get abs value */
	if (delta < 0) delta = - delta;
	/* if delta < 20% of whole then ok return 1 */
	if (delta < tw) return 1;
	return 0;
};


/* return 1 random byte from RNG
*  There are 3 cases
*
*	1) hardrandomdriver is defined (i.e. not "")
*		in this case we open hardrandomdriver as a file
*		with fopen. Each random bytes is read directly
*		from this file.
*	2) Hardrandomdriver is not defined but hardrandom_port is (i.e. not=0)
*	  a)	OS/2 system open the $TESTCFG driver
*		use special calls to this driver to read bytes directly from
*		the port.
*	  b)	not OS/2 ie MSDOS
*		read random bytes directly from the port with "IN"
*		machine instruction.
*/
byte hardrand(void)
{
    /* if our random driver can be opened as a file. */
    if ( *hardrandomdriver )
    {
	static FILE *rand=NULL;			/* file to RNG */
	if (rand == NULL)			/* if not opened */
	{
	  /* open the RNG as a file. We are assuming that
	  * the device gives one random byte each time it is read.
	  */
	  rand =fopen(hardrandomdriver,"rb");
	  if (rand == NULL)
	  {
		fprintf(pgpout,
   		 LANG("\n\007Can't open file '%s'\n"),hardrandomdriver);
	  	exitPGP(INVALID_FILE_ERROR);
	  };
	  setvbuf(rand,NULL,_IONBF,64);
	  /* set small buffer driver may be slow and we
	  probably do not need large numbers of random bytes */

	  /* test that RNG is really random */
	  if ( !RNGgood() )
	  {
		fprintf(pgpout,
   		 LANG("\n\007Can't open file '%s'\n"),hardrandomdriver);
	  	exitPGP(INVALID_FILE_ERROR);
	  };
	};
	/* return one random byte from file.*/
	return  (fgetc(rand) & 0xff);
    }
#if defined(MSDOS) || defined(__MSDOS__)
/* the dos os can get bytes directly from a bus RNG.
 * if hardrandom port =RNGPORT is defined in config.sys
 * then try to get bytes from the port.
 */
    if (hardrandom_port)
    {
        static firstused=0;		/* first time flag */
        static char filename[25];	/* holds filename for error */
        if (firstused == 0 )		/* if first call */
        {
          firstused = 1;		/* no longer first call */
          /* create phony file name for error messages.*/
          sprintf(filename,"PORT=%x",hardrandom_port);
          /* test for real randomness, if not abort */
	  if ( !RNGgood() )
	  {
		fprintf(pgpout,
   		 LANG("\n\007Can't open file '%s'\n"),"TESTCFG$");
	  	exitPGP(INVALID_FILE_ERROR);
	  };
        };
#ifdef RANDHARDWAIT
	/* must wait between accessing random no generator.
	*  must use spin lock because OS'es do not provide
	*  fine enough timer resolution.
	*  it is possible that this code could cause an unnecessary
	*  wait. The probability of such a wait is so small, and
	*  the wait is so short that this is judged not important.
	*/
	{
	 static unsigned short old=0;
	 unsigned short NewTime;
	 do {
	  outp(timercntl,timer0rwl);	/*latch timer */
	  					/* get current time */
	  * ( (char *) &NewTime) = inp(timer0);
	  * ( ( (char *) &NewTime )+1) = inp(timer0);
	 } while ( (unsigned short)
		( (signed short)NewTime - (signed short) old ) < RANDHARDWAIT);
		/* loop untill delta is greater than RANDHARDWAIT */

	 old =NewTime;		/* SAVE current time for next call */

	}
#endif /* RANDHARDWAIT */
	/* return one random byte from the bus RNG */
	return (inp(hardrandom_port) & 0xff);
    };
#endif
/* if we are using OS/2 with a bus RNG, we can not access the
 * hardware directly. Instead we use IBM's TESTCFG driver to
 * do our in instructions for us.
 */
#if defined(OS2) || defined(__OS2__)
/* if hardrandom_port = RNGPORT from config.txt is
 * defined, we get random bytes directly from the port
 */
    if (hardrandom_port)
    {
        static char filename[25];	/* bogus filename for the port */
	/* OS/2 return code.*/
	APIRET rc;
	/* handle to TESTCFG driver. */
	static HFILE handle=0;
	/* action taken is returned by DosOpen */
	ULONG ActionTaken;

	ULONG portsiz;	/*size of parameter block */

	ULONG data;	/* data block */
	ULONG datasiz;	/* size of data block */

	/* parameter block. to get byte from port */
	struct inio {
		USHORT 	io_address;
		USHORT	data_width;
	} port ;

	/* address of io port. */
	port.io_address = hardrandom_port;
	/* byte has be gotten. */
	port.data_width = 1;

	/* if we have not opened TESTCFG$, do it now */
	if (handle == 0)
	{
          sprintf(filename,"PORT=%x",hardrandom_port);
	  rc = DosOpen("TESTCFG$",&handle,&ActionTaken,0L,
		FILE_NORMAL,
		(OPEN_ACTION_CREATE_IF_NEW|OPEN_ACTION_OPEN_IF_EXISTS),
		(OPEN_ACCESS_READONLY|OPEN_SHARE_DENYNONE),
		0L);

	  if (rc)	/* unable to open TESTCFG driver. */
	  {
		fprintf(pgpout,
   		 LANG("\n\007Can't open file '%s'\n"),"TESTCFG$");
	  	exitPGP(INVALID_FILE_ERROR);
	  };

	  if ( !RNGgood() )	/* test the device is truely random */
	  {
		fprintf(pgpout,
   		 LANG("\n\007Can't open file '%s'\n"),filename);
	  	exitPGP(INVALID_FILE_ERROR);
	  };
	}

	/* size of parameter block on input */
	portsiz=sizeof(port);

	/* 0 bytes of data block input */
	datasiz=0;

	/* category 80h, function 41h, does a in ins on the port. */
	rc = DosDevIOCtl(handle,0x80,0x41,&port,sizeof(port),&portsiz,
		&data,sizeof(data),&datasiz);

	if (rc)		/* unable to read the port */
	{
		fprintf(pgpout,
   		 LANG("\n\007Can't open file '%s'\n"),"TESTCFG$");
	  	exitPGP(INVALID_FILE_ERROR);
	};
	/* return one byte read. */
	return (data & 0XFF);
    };
#endif
return 0;
};
